package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * EllipsoidConvolutionKernel.java
 *
 * @author Created by G Odell, june 2006
 */
import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;

/* when tiltangle==0, the formula for the boundary of the ellipsoidal ball is
(x - xctr)^2/r1^2 + (y - yctr)^2/r2^2 == 1

multiply by (r1*r2)^2
((x - xctr)*r2)^2  + ((y - yctr)*r1)^2  == (r1*r2*)^2
how many ellipsoids to consider. Let r1 range from 2 to maxrad and let r2 range from 2 to r1. 
The number of r1 - r2 combinations is then Nrads = 1+(1/2)(maxrad-2)^2
maxrad  Nrads
2 		1
3 		2
4 		3
5 		6	
6 		9
7 		14
8 		19
9 		26
10 		33
11 		42
12 		51
13 		62
14 		73
15 		86
16 		99

any ellipsoid can have any tilt spanning 180 degrees. If I do this in 2 degree increments, there are
90 different tilts. This makes 90,000 possible convolution kernels to compute and store. 
The average number of x-y points in a typical ellipsoid will be something like 200, with one point occupying three bytes.
This adds up to about 3*200*90000 = 27 MB. But every is mirror symmetric so we need only stor half the pattern. 
Thus we need 14 MB of storage. Not too much.

if (ix,iy) are the offsets stored, the rotated offsets are:

ixrot ->  costilt*ix + sintilt*iy
iyrot -> -sintilt*ix + costilt*iy
If save only the unrotated ellipsoids, then that takes only 90000*3 = 270,000 bytes. Good. 
*/	

public class EllipsoidConvolutionKernel extends KernelConvolver{

static int MAXKERNELRAD = 50; // max allowable value for major or minor axis
static EllipsoidConvolutionKernel [][] allEllipsiodalKernels = new EllipsoidConvolutionKernel [1+MAXKERNELRAD][1+MAXKERNELRAD]; 
// constructs and saves kernels as they are needed. Use only the entries on and above the diagonal. 
double rad1,rad2;
public void sepuku() { // frees all storage
	allEllipsiodalKernels=null;
	}

	
static public EllipsoidConvolutionKernel getEllipsoidalKernel(int r1, int r2) {
	if(r1 >= MAXKERNELRAD || r2 >= MAXKERNELRAD) {
		talk("getEllipsoidalKernel() called with one radius > "+MAXKERNELRAD);
		double [] kpars = {r1, r2, 1.6}; // make one up on the spot but do not save it.
		return new EllipsoidConvolutionKernel(kpars);
		}
		
	if(r1 < 1 ||  r2 < 1) {
		talk("Ellipsoidal kernel must have both radii > 0");
		return null;
		}
	if(r1 > r2) {
		int rr=r2;
		r2=r1;
		r1=rr;
		}
	if(allEllipsiodalKernels[r1][r2] == null) {
		double [] kpars = {r1, r2, 1.6};
		allEllipsiodalKernels[r1][r2] = new EllipsoidConvolutionKernel(kpars);
		}
	return allEllipsiodalKernels[r1][r2];
	}
	
public EllipsoidConvolutionKernel(double [] kernelPars) {
	super(kernelPars);
	}

public int getAnglemin() { // returns the starting rotation angle for fullSpinAndExploreCenterConvolution(); overwrite for each extension
	if(rad1==rad2)
		return 0; // supresses spin
	else
		return -90;
	}
	
public int getAnglemax() { // returns the ending rotation angle for fullSpinAndExploreCenterConvolution(); overwrite for each extension
	if(rad1==rad2)
		return 0; // supresses spin
	else
		return 90;
	}

public double getAngleStep() { // returns the angle increment forfullSpinAndExploreCenterConvolution() to use ; overwrite for each extension
	return 30;
	}
	
	
public void makeKernel() {
// this example, to be overwritten, defines a circular kernel of radius kernelPars[0] and
// rim width kernelPars[1]
// equation for ellipsoid interior ((x - xctr)*r2)^2  + ((y - yctr)*r1)^2  <= (r1*r2*)^2
	int x,y,j;
	double rad1WithRim =  kpars[0] + kpars[2], rad2WithRim = kpars[1] + kpars[2];
	interiorXYoffsets=new short[1][1];
	boundaryOffsets=new short[1][1];
	nInteriorPoints=0;
	nBoundaryPoints=0;
	tiltangle=0;
	rad1 = kpars[0];
	rad2 = kpars[1];

	double xyrfac = (rad1*rad2*rad1*rad2),xr,yr;
	double xyrimfac = (rad1WithRim*rad2WithRim*rad1WithRim*rad2WithRim);
	talk("getEllipsoidalKernel.makeKernel() called for r1="+kpars[0]+" and r2 = "+kpars[1]);
	for(j=0;j<2;j++) { // on j==0 pass, just count the points; store the offsets on j==1 pass
		for(x=-(int)(1.5*rad1WithRim); x<=1.5*rad1WithRim; x++)
			for(y=-(int)rad2WithRim; y<=rad2WithRim; y++) {
				xr = x*rad2WithRim;
				yr = y*rad1WithRim;
				if(xr*xr + yr*yr <= xyrimfac) { // the point is in the rim or in the interior
					xr = x*rad2;
					yr = y*rad1;
					if(xr*xr + yr*yr <= xyrfac) { // the point is in  the interior
						if(j>0) {
							interiorXYoffsets[nInteriorPoints][0] = (short)x;	
							interiorXYoffsets[nInteriorPoints][1] = (short)y;
							}	
						++nInteriorPoints;
						}
					else { // the point is in  the rim
						if(j>0) {
							boundaryOffsets[nBoundaryPoints][0] = (short)x;	
							boundaryOffsets[nBoundaryPoints][1] = (short)y;
							}	
						++nBoundaryPoints;
						}
					}
				}

			if(j==0) {
				interiorXYoffsets = new short[1+nInteriorPoints][2];
				boundaryOffsets = new short[1+nBoundaryPoints][2];
				talk("EllipsoidConvolutionKernel.makeKernel() makes offsets arrays with "+nInteriorPoints+" and "+nBoundaryPoints+" points");
				nInteriorPoints=0;
				nBoundaryPoints=0;
				}
			
		}
	}

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
		
}

