package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;



/**
 * KernelConvolver.java
 *
 * @author Created by Omnicore G Odell - june 2006
  * This class holds a 2-D raster of values defining a pattern to be convoolved with an image. 
   *The Kernel comprises two lists of x-y offsets, from the center of the kernel, of points in the kernel.
    *The first list has offsets for the pattern points that are 'ON' and the second list has offsets for
     *points that are 'OFF'. The intent is that the 'OFF' points define a continuous border
      * of the kernel (which might be a one-pixel-wide curve bounding the kernel. The boundary can be null 
       * if no points are wanted in it.
   *The class has methods for convolving the kernel with an image after shifting the center of the 
    *kernel, and rotating the kernel through a specified angle. These methods return either the convolution
     *score of the convolution score normalized by the number of pixels in the kernel. These methods need
      *not be overwritten. 
   *There are higher order methods for convoloving the kernel with an image which explore specified
    *ranges of center-shifts and rotation angles and then return the best offset and best rotation angle along
     *with the convolution score. 
   * this class has methods for making the kernel which must be overwritten to shape specific kernels. 
    *The class constructors call these kernel-making methods just once.  
     *
   * To make computing kernel convolutions go faster, this class (to be extended by classes that define and
    *manipulate specific kernels) holds the image to be convolved in a static field that all extensions of  
     *KernelConvolver operate on. 
      *
   * This basic class, intended to be extended, illustrates how to use it for circular convolution kernels.         
 */
public class KernelConvolver  {
public int nInteriorPoints=0, nBoundaryPoints=0;
public short [][] interiorXYoffsets=null, boundaryOffsets=null;
public double tiltangle;
public double [] kpars=null;
// rad1,rad2,rad3,rad4 are parameters that define the shape of the kernel. They may or may not be radii.

public short [][] rotatedInteriorXYoffsets=null, rotatedBoundaryOffsets=null;
public String nameOfKernel=null;

static short [] theShortRaster = null;
static int nrows=0, ncols=0;

public void rotateKernel(double angle) {
	if(nInteriorPoints <= 0)
		return;
	angle = angle*2*Math.PI/360.;
	tiltangle = angle;
	double cosang = Math.cos(angle);
	double sinang = Math.sin(angle);
	
	// this is tricky because I should work backward using the inverse mapping. But I
	// don't know what rectaangle the rotated kernel occupies.
	int i;
	for(i=0; i<nInteriorPoints; i++) {
		 	rotatedInteriorXYoffsets[i][0] = (short)(.4999 + cosang*interiorXYoffsets[i][0] - sinang*interiorXYoffsets[i][1]);
			rotatedInteriorXYoffsets[i][1] = (short)(.4999 + sinang*interiorXYoffsets[i][0] + cosang*interiorXYoffsets[i][1]);			
			}
	for(i=0; i<nBoundaryPoints; i++) {
		 	rotatedBoundaryOffsets[i][0] = (short)(.4999 + cosang*boundaryOffsets[i][0] - sinang*boundaryOffsets[i][1]);
			rotatedBoundaryOffsets[i][1] = (short)(.4999 + sinang*boundaryOffsets[i][0] + cosang*boundaryOffsets[i][1]);			
			}
	}

public int getAnglemin() { // returns the starting rotation angle for fullSpinAndExploreCenterConvolution(); overwrite for each extension
	return 0;
	}
	
public int getAnglemax() { // returns the ebding rotation angle for fullSpinAndExploreCenterConvolution(); overwrite for each extension
	return 0;
	}

public double getAngleStep() { // returns the angle increment forfullSpinAndExploreCenterConvolution() to use ; overwrite for each extension
	return 3;
	}
	
public double [] fullSpinAndExploreCenterConvolution(int xcen, int ycen, double explorerad) {
// for each angle, rotate the kernel, and move (xcen,ycen) around through the full range within explorerad of (xcen,ycen). 
// keep track of best score obtained.
// returns {score,xcen,ycen,rad1,rad2,angle}
	int xc,yc,delx,dely;
	double aScore;
	int anglemin = getAnglemin(), anglemax = getAnglemax();
	double anglestep = getAngleStep();	
	int xcenmin = (int)(xcen-explorerad-.49);
	int xcenmax = (int)(xcen+explorerad+.49);
	int ycenmin = (int)(ycen-explorerad-.49);
	int ycenmax = (int)(ycen+explorerad+.49);
	double exploreradsq = explorerad*explorerad;
	double [] bestScoreXYoffsRot8 = {77777.E25, xcen,ycen,anglemin};// = {score,xcen,ycen,rot8}
	while(anglemin <= anglemax) {
		rotateKernel(anglemin);
		for(xc = xcenmin; xc <= xcenmax; ++xc)
			for(yc = ycenmin; yc <= ycenmax; ++yc) {
				delx = xc-xcen;
				dely = yc-ycen;				
				if(delx*delx+dely*dely<=exploreradsq && (aScore = convolveRotatedKernelTranslatedTo(xc, yc)) < bestScoreXYoffsRot8[0]) {
					bestScoreXYoffsRot8[0] = aScore;
					bestScoreXYoffsRot8[1] = xc;
					bestScoreXYoffsRot8[2] = yc;
					bestScoreXYoffsRot8[3] = anglemin;
					}
				}
		anglemin += anglestep;
		}		
	return bestScoreXYoffsRot8;
	}

static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
			
public double convolveRotatedKernelTranslatedTo(int xcen, int ycen) {  
	if(nInteriorPoints <= 0 || nBoundaryPoints <= 0)
		return -77777;
	long interiorscore=0,boundaryscore=0;
	int i,ix,iy;
		
	for(i=0; i<nInteriorPoints; i++) {
		if((ix =  xcen + rotatedInteriorXYoffsets[i][0]) >= 0 &&
			ix < ncols && 
				(iy = ycen + rotatedInteriorXYoffsets[i][1]) >= 0 &&
					iy < nrows) 
						interiorscore += theShortRaster[iy*ncols + ix];
		}

	for(i=0; i<nBoundaryPoints; i++) {
		if((ix =  xcen + rotatedBoundaryOffsets[i][0]) >= 0 &&
			ix < ncols && 
				(iy = ycen + rotatedBoundaryOffsets[i][1]) >= 0 &&
					iy < nrows) 
						boundaryscore += theShortRaster[iy*ncols + ix];
			}
//	talk("score="+(interiorscore - boundaryscore)/(double)interiorscore+" from KernelConvolver.convolveRotatedKernelTranslatedTo("+xcen+", "+ycen+") r1="+kpars[0]+", rad2="+kpars[1]+", angle="+tiltangle);
	return (interiorscore - boundaryscore)/(double)interiorscore;	
	}

public static void setShortRaster(int nc, int nr, short [] raster) {
	ncols = nc;
	nrows = nr;
	theShortRaster = raster;
	}
	
public void makeKernel() {
// this example, to be overwritten, defines a circular kernel of radius kernelPars[0] and
// rim width kernelPars[1]
	nInteriorPoints=0;
	nBoundaryPoints=0;
	double rad1 = kpars[0]; // kpars has been stored already by the constructor
	int x,y,j,squareSideHalfWidth = (int)(.49 + kpars[0] + kpars[1]);
	double rsq,radsq=kpars[0]*kpars[0], rimradsq = kpars[0] + kpars[1];
	rimradsq = rimradsq*rimradsq;
	interiorXYoffsets=new short[1][1];
	boundaryOffsets=new short[1][1];
	for(j=0;j<2;j++) { // on j==0 pass, just count the points; store the offsets on j==1 pass
		for(x=-squareSideHalfWidth; x<=squareSideHalfWidth; x++)
			for(y=-squareSideHalfWidth; y<=squareSideHalfWidth; y++) {
				if( (rsq = x*x + y*y) <= rimradsq) {
					if(rsq <= radsq) {
						if(j>0) {
							interiorXYoffsets[nInteriorPoints][0] = (short)x;	
							interiorXYoffsets[nInteriorPoints][1] = (short)y;
							}	
						++nInteriorPoints;
						}
					}
				else { // must be in rim
						if(j>0) {
							boundaryOffsets[nBoundaryPoints][0] = (short)x;	
							boundaryOffsets[nBoundaryPoints][1] = (short)y;
							}	
						++nBoundaryPoints;
					}
		if(j==0) {
			interiorXYoffsets = new short[nInteriorPoints][2];
			boundaryOffsets = new short[nInteriorPoints][2];;
			nInteriorPoints=0;
			nBoundaryPoints=0;
			}
		}
	}}

public KernelConvolver(double [] kernelPars) {
	int npars = kernelPars.length;
	kpars = new double[npars]; // make a copy of kernelPars
	for(int i=0;i<npars;i++)
		kpars[i]=kernelPars[i];
	makeKernel();// each extension overwrites this method.	
	nInteriorPoints = interiorXYoffsets.length; 
	nBoundaryPoints = boundaryOffsets.length;
 	rotatedInteriorXYoffsets = new short[nInteriorPoints][2]; // make space for rotated versions of this kernel
 	rotatedBoundaryOffsets = new short[nBoundaryPoints][2];
	}

}

