package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class SkelFltr {
	SegmenterNode nd=null;
	short [] work=null;
	int workrows=-1, workcols=-1;
	int [] skeloffs = {-4,-3,-2,-1,1,2,3,4};
	int [] twoto = {1,2,4,8,16,32,64,128};
	boolean [] blacktable = new boolean[256];
	boolean ifVirgin=true;
	
	public SkelFltr(SegmenterNode nd) {
		this.nd=nd;
		}
	
	private void makeWork(int incols, int inrows) {
		if(work==null || incols != workrows || inrows != workrows) {
			workrows=inrows;
			workcols=incols;
			work = new short[inrows*incols];
			}
		}
		
	public void tellWhatYouDo() {
		nd.talk("");
		nd.talk("The SkelFltr filter has a single input raster and two parameters, 'minPix' and 'iter8'");
		nd.talk("It performs iter8 successive iterations of the following procedure");
		nd.talk("It erases pixels which are on but have certain patterns of on neighbors which");
		nd.talk("mark those pixels as being on the outer edge. The patterns causing a central 'on'");
		nd.talk("pixel to be turned off are the following ones, plus all their rotations by 90, 180,");
		nd.talk("and 270 degrees. Plus their mirror flips and those rotations of the flips");
		nd.talk("                     The basis patterns are");
		nd.talk("          111  111   111   110  010  100  110");
		nd.talk("          0X0  1X0   0X1   1X0  0X0  0X0  0X0");
		nd.talk("          000  000   001   000  000  000  000");
		nd.talk("These patterns lead to a filter which eats away at edges but will not break");
		nd.talk("through a continuous smooth line of pixels");
		nd.talk("This filter quits whenever one iteration through it changes fewer than .0125% of the pixels.");
		nd.talk("Thus, to get this filter to skeletonize until it stops changing many pixels, set iter8 to");
		nd.talk("some high value.");
		}
		
  public boolean doit(short [] inras, int incols, int inrows, int thresh, int iter8, short [] outras) {
	int i,j,npix=incols*inrows,howmanychanged=0;
	String stg = "SkelFltr turned off (";
	if(iter8<=0) 
		iter8=1; // must do at least one iteration

	
	if(ifVirgin) {
		ifVirgin=false;
		makeBlackTable(incols);
		}
	
	// create the correct offset table for the input raster 
	skeloffs[0] = (-1) -incols;     skeloffs[1] = skeloffs[0]+1;     skeloffs[2] = skeloffs[1]+1;
	skeloffs[3] = (-1);                                              skeloffs[4] = 1;
	skeloffs[5] = incols -1;        skeloffs[6] = skeloffs[5]+1;     skeloffs[7] = skeloffs[6]+1;
	
	makeWork(incols,inrows); // make the work array which erodeOnce uses.

	skelthreshold(inras,outras,thresh,incols,inrows); // this does a black/white thresholding on input,
												// putting the result in outras.
	while(iter8>0) {
		i =  erodeOnce(outras, incols, inrows);
		howmanychanged += i;
		stg = stg + (i+" + ");
		if(i < npix/8000)  
			iter8=0; // we are not making worthwhile progress, so stop the iteration
		else --iter8;
		}
	
	nd.talk(stg+") = "+howmanychanged+" pixels.");
	return(true);
	}

	private int neighborPatternCode(int offset, short [] ras) {
/* There are 2^8 possible n-off patterns of nearest neighbors to a central pixel (which is on), and
this method returns the unique code for the pattern of neighbors the pixel at offset has. 
Blacktable keeps track of which patterns, i.e. which codes, should have the central pixel turned off.  */
		int code=0;
		for(int j=0;j<8;j++) 
			if(ras[offset+skeloffs[j]] > 127)
				code +=twoto[j];
		return code;		
		}
		
		
	private void makeBlackTable(int incols)  {
	int i,j,code,howmany=0;
/* These are the basis patterns that cause the central pixel, X,
 to go to black. A 'no' above the letter means we don't use it     (as of 2/2/2001)
          no        no        no   no   no                
 a    b    c    d    e    f    g    h    i    j    k   l
111  111  111  111  111  110  111  110  000  010  100  110
0X0  1X0  1X1  0X1  1X0  1X0  1X1  1X1  0X0  0X0  0X0  0X0
000  000  000  001  110  000  100  000  000  000  000  000
 4    5    6    6    7    4    7    5    1    2    2    3  <--- # pixels on
		 */
	short [][] abcdefghijkl = {{255,255,255,   0,255,0,      0,0,0},      // a
								{255,255,255,   255,255,0,    0,0,0},     // b
							//	{255,255,255,   255,255,255,  0,0,0},     // c
								{255,255,255,   0,255,255,    0,0,255},   // d
							//	{255,255,255,   255,255,0,    255,255,0}, // e
								{255,255,0,     255,255,0,    0,0,0},     // f
							//	{255,255,255,   255,255,255,  255,0,0},   // g
							//	{255,255,0,     255,255,255,  0,0,0},     // h
							//	{0,0,0,         0,255,0,      0,0,0},     // i
								{0,255,0,       0,255,0,      0,0,0},     // j
								{255,0,0,       0,255,0,      0,0,0},     // k
								{255,255,0,     0,255,0,      0,0,0}};    // l 
	int nBasisPats = 7;

/* apply these transformations: vflip, rotate90, rotate180 rotate-90
  
012	678		258	876	630   036  210  852
345	345		147	543	741   147  543  741
678	012		036	210	852   258  876  630

std	flip    r90 r180 r-90 r90  r180 r-90   
            std std  std  flip flip flip 
*/
	int [][] permutes ={{0,1,2,3,4,5,6,7,8}, // standard order
						{6,7,8,3,4,5,0,1,2}, // flip over horizontal axis
						{2,5,8,1,4,7,0,3,6}, // rotate std 90 degrees counter clockwise 
						{8,7,6,5,4,3,2,1,0}, // rotate std 180 degrees
						{6,3,0,7,4,1,8,5,2}, // rotate std 90 degrees clockwise
						{0,3,6,1,4,7,2,5,8}, // rotate flip 90 degrees counter clockwise
						{2,1,0,5,4,3,8,7,6}, // rotate flip 180 degrees
						{8,5,2,7,4,1,6,3,0}};// rotate flip 90 degrees clockwise

	// create offset array for the 3x3 rasters in abcdefghijkl arrays above
	skeloffs[0] = -4;
	skeloffs[1] = -3;
	skeloffs[2] = -2;

	skeloffs[3] = -1;
	skeloffs[4] = 1;

	skeloffs[5] = 2;
	skeloffs[6] = 3;
	skeloffs[7] = 4; 

		for(i=0;i<256;i++)
			blacktable[i]=false;  // means these patterns don't get erased unless next loop turns them true
		
		for(i=0;i<nBasisPats;i++)
			for(j=0;j<8;j++) {
				code=neighborPatternCode(4, rot8(abcdefghijkl[i], permutes[j]));
				if(!blacktable[code]) {
					blacktable[code] = true;
					++howmany;
					}
				}
			nd.talk("SkelFltr turns off central pixel in "+howmany+" out of 256 possible neighbor patterns");
	}
	
	private short [] rot8(short [] square, int [] perm) {
		short [] retn = new short[9];
		int i;
		for(i=0;i<9;i++)
			retn[i] = square[perm[i]];
		return retn;
		}
	
	private int erodeOnce(short [] ras, int NCOLS, int NROWS) { // does one pass of erroding. 
			// uses the work array
			// returns the number of pixeels turned off
		int npix = NCOLS*NROWS;
		int i,row,col,offset,code,numTurnedOff=0;
		for(i=0;i<npix;i++)
			work[i]=ras[i]; // copy input array to output
		for(row=1;row<NROWS-1;row++) {
			offset = row*NCOLS;
			for(col=1;col<NCOLS-1;col++) {
				++offset;
				if(work[offset]>127) { // only if pixeel is on do we see if we should turn it off
					code = neighborPatternCode(offset, ras);
					if(blacktable[code]) {
						work[offset]=0;
						++numTurnedOff;
						}
 					}
 				}
 			}
 		for(i=0;i<npix;i++)
 			ras[i]=work[i];
 		return numTurnedOff;
		}	

/****************** skelthreshold() ***********/
private void skelthreshold(short [] in,short [] out,int thresh,int NCOLS,int NROWS) {
	int npix=NCOLS*NROWS,i;
	short on = (short )255;
	
	for(i=0;i<npix;i++) 
  		out[i] = in[i]>thresh? on : 0;
	return;
}

/****************** thikn() ***********/
private int thikn(short [] ras,short fill,int NCOLS,int NROWS) { // turn on any pixel whose neighbor is on
	int i,npix=NCOLS*NROWS,j,k,nfilled,offset; // uses work array
	for(i=0;i<npix;i++) work[i]=ras[i];
	nfilled=0;

	for(i=1;i<NROWS-1;i++) {
		offset = i*NCOLS;
		for(j=1;j<NCOLS-1;j++) {
			++offset;
			if (ras[offset]==0) { /* the pixel starts out dark */
				k=0;
				while(k<8) {
					if(ras[skeloffs[k]+offset]>0) {
						work[offset]=fill;
						++nfilled;
						k=9;
						}
					else ++k;
					}
				}
			}
		}
		
	for(i=0;i<npix;i++) 
		ras[i]=work[i];
	return(nfilled);
	}
	

}